@@ -8,7 +8,7 @@
  * THE GNU PUBLIC LICENSE 2, WHICH IS INCLUDED WITH THIS SOURCE.    *
  * PLEASE READ THESE TERMS BEFORE DISTRIBUTING.                     *
  *                                                                  *
- * THE Ogg123 SOURCE CODE IS (C) COPYRIGHT 2000-2001                *
+ * THE Ogg123 SOURCE CODE IS (C) COPYRIGHT 2000-2005                *
  * by Stan Seibert <volsung@xiph.org> AND OTHER CONTRIBUTORS        *
  * http://www.xiph.org/                                             *
  *                                                                  *
@@ -18,6 +18,10 @@
 
  ********************************************************************/
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
+
 #include <sys/types.h>
 #include <string.h>
 #include <stdlib.h>
@@ -42,16 +46,19 @@
 #include "status.h"
 #include "playlist.h"
 #include "compat.h"
+#include "remote.h"
 
 #include "ogg123.h"
 #include "i18n.h"
 
+extern int exit_status; /* from status.c */
 
 void exit_cleanup ();
 void play (char *source_string);
 
+#define PRIMAGIC (2*2*2*2*3*3*3*5*7)
 /* take buffer out of the data segment, not the stack */
-#define AUDIO_CHUNK_SIZE 4096
+#define AUDIO_CHUNK_SIZE ((32768 + PRIMAGIC - 1)/ PRIMAGIC * PRIMAGIC)
 unsigned char convbuffer[AUDIO_CHUNK_SIZE];
 int convsize = AUDIO_CHUNK_SIZE;
 
@@ -76,6 +83,8 @@ file_option_t file_opts[] = {
    &options.default_device, NULL},
   {0, "shuffle",        N_("shuffle playlist"),      opt_type_bool,
    &options.shuffle,        &int_0},
+  {0, "repeat",         N_("repeat playlist forever"),   opt_type_bool,
+   &options.repeat,         &int_0}, 
   {0, NULL,             NULL,                    0,               NULL,                NULL}
 };
 
@@ -143,8 +152,9 @@ void options_init (ogg123_options_t *opts)
   opts->delay = 500;
   opts->nth = 1;
   opts->ntimes = 1;
-  opts->seekpos = 0.0;
+  opts->seekoff = 0.0;
   opts->endpos = -1.0; /* Mark as unset */
+  opts->seekmode = DECODER_SEEK_NONE;
   opts->buffer_size = 128 * 1024;
   opts->prebuffer = 0.0f;
   opts->input_buffer_size = 64 * 1024;
@@ -153,10 +163,86 @@ void options_init (ogg123_options_t *opts)
 
   opts->status_freq = 10.0;
   opts->playlist = NULL;
+  opts->remote = 0;
+  opts->repeat = 0;
 
 }
 
+double strtotime(char *s)
+{
+	double time;
 
+	time = strtod(s, &s);
+
+	while (*s == ':')
+		time = 60 * time + strtod(s + 1, &s);
+
+	return time;
+}
+
+void set_seek_opt(ogg123_options_t *ogg123_opts, char *buf) {
+
+  char *b = buf;
+
+  /* skip spaces */
+  while (*b && (*b == ' ')) b++;
+  
+  if (*b == '-') {
+  /* relative seek back */
+    ogg123_opts->seekoff = -1 * strtotime(b+1);
+    ogg123_opts->seekmode = DECODER_SEEK_CUR;
+  } else
+  if (*b == '+') {
+  /* relative seek forward */
+    ogg123_opts->seekoff = strtotime(b+1);
+    ogg123_opts->seekmode = DECODER_SEEK_CUR;
+  } else {
+  /* absolute seek */
+    ogg123_opts->seekoff = strtotime(b);
+    ogg123_opts->seekmode = DECODER_SEEK_START;
+  }
+}
+
+int handle_seek_opt(ogg123_options_t *options, decoder_t *decoder, format_t *format) {
+
+  float pos=decoder->format->statistics(decoder)->current_time;
+
+  /* this functions handles a seek request. It prevents seeking out
+     of band, i.e. before the beginning or after the end. Instead,
+	 it seeks to the start or near-end resp. */
+
+  if (options->seekmode != DECODER_SEEK_NONE) {
+
+    if (options->seekmode == DECODER_SEEK_START) {
+      pos = options->seekoff;
+    } else {
+      pos += options->seekoff;
+    }
+
+    if (pos < 0) {
+      pos = 0;
+    }
+
+    if (pos > decoder->format->statistics(decoder)->total_time) {
+      /* seek to almost the end of the stream */
+      pos = decoder->format->statistics(decoder)->total_time - 0.01;
+    }
+
+    if (!format->seek(decoder, pos, DECODER_SEEK_START)) {
+      status_error(_("Could not skip to %f in audio stream."), options->seekoff);
+#if 0
+      /* Handle this fatally -- kill the audio thread */
+      if (audio_buffer != NULL)
+	buffer_thread_kill(audio_buffer);
+#endif
+    }
+  }
+
+  options->seekmode = DECODER_SEEK_NONE;
+  
+  return 1;
+}
+  
 /* This function selects which statistics to display for our
    particular configuration.  This does not have anything to do with
    verbosity, but rather with which stats make sense to display. */
@@ -184,6 +270,7 @@ void select_stats (stat_format_t *stats, ogg123_options_t *opts,
     stats[6].enabled = 0;
     stats[7].enabled = 0;
   }
+  free(data_source_stats);
     
   /* Assume we need total time display, and let display_statistics()
      determine at what point it should be turned off during playback */
@@ -206,38 +293,48 @@ void display_statistics (stat_format_t *stat_format,
 					source->transport->statistics(source),
 					decoder->format->statistics(decoder));
 
-  /* Disable/Enable statistics as needed */
+  if (options.remote) {
+  
+    /* Display statistics via the remote interface */
+    remote_time(pstats_arg->decoder_statistics->current_time, 
+                pstats_arg->decoder_statistics->total_time,
+				pstats_arg->decoder_statistics->instant_bitrate);
+				
+  } else {
+  
+	/* Disable/Enable statistics as needed */
 
-  if (pstats_arg->decoder_statistics->total_time <
-      pstats_arg->decoder_statistics->current_time) {
-    stat_format[2].enabled = 0;  /* Remaining playback time */
-    stat_format[3].enabled = 0;  /* Total playback time */
+	if (pstats_arg->decoder_statistics->total_time <
+    	pstats_arg->decoder_statistics->current_time) {
+      stat_format[2].enabled = 0;  /* Remaining playback time */
+      stat_format[3].enabled = 0;  /* Total playback time */
+	}
+
+	if (pstats_arg->data_source_statistics->input_buffer_used) {
+      stat_format[6].enabled = 1;  /* Input buffer fill % */
+      stat_format[7].enabled = 1;  /* Input buffer state  */
+	}
+
+	if (audio_buffer) {
+      /* Place a status update into the buffer */
+      buffer_append_action_at_end(audio_buffer,
+				  &print_statistics_action,
+				  pstats_arg);
+
+      /* And if we are not playing right now, do an immediate
+    	 update just the output buffer */
+      buffer_stats = buffer_statistics(audio_buffer);
+      if (buffer_stats->paused || buffer_stats->prebuffering) {
+    	pstats_arg = new_print_statistics_arg(stat_format,
+					      NULL,
+					      NULL);
+    	print_statistics_action(audio_buffer, pstats_arg);
+      }
+      free(buffer_stats);
+
+	} else
+      print_statistics_action(NULL, pstats_arg);
   }
-
-  if (pstats_arg->data_source_statistics->input_buffer_used) {
-    stat_format[6].enabled = 1;  /* Input buffer fill % */
-    stat_format[7].enabled = 1;  /* Input buffer state  */
-  }
-
-  if (audio_buffer) {
-    /* Place a status update into the buffer */
-    buffer_append_action_at_end(audio_buffer,
-				&print_statistics_action,
-				pstats_arg);
-    
-    /* And if we are not playing right now, do an immediate
-       update just the output buffer */
-    buffer_stats = buffer_statistics(audio_buffer);
-    if (buffer_stats->paused || buffer_stats->prebuffering) {
-      pstats_arg = new_print_statistics_arg(stat_format,
-					    NULL,
-					    NULL);
-      print_statistics_action(audio_buffer, pstats_arg);
-    }
-    free(buffer_stats);
-    
-  } else
-    print_statistics_action(NULL, pstats_arg);
 }
 
 
@@ -346,13 +443,15 @@ int main(int argc, char **argv)
   }
 
   /* Don't use status_message until after this point! */
-  status_set_verbosity(options.verbosity);
+  status_init(options.verbosity);
 
   print_audio_devices_info(options.devices);
 
   
   /* Setup buffer */ 
   if (options.buffer_size > 0) {
+    /* Keep sample size alignment for surround sound with up to 10 channels */
+    options.buffer_size = (options.buffer_size + PRIMAGIC - 1) / PRIMAGIC * PRIMAGIC;
     audio_buffer = buffer_create(options.buffer_size,
 				 options.buffer_size * options.prebuffer / 100,
 				 audio_play_callback, &audio_play_arg,
@@ -365,20 +464,6 @@ int main(int argc, char **argv)
     audio_buffer = NULL;
 
 
-  /* Shuffle playlist */
-  if (options.shuffle) {
-    int i;
-    
-    srandom(time(NULL));
-    
-    for (i = 0; i < items; i++) {
-      int j = i + random() % (items - i);
-      char *temp = playlist_array[i];
-      playlist_array[i] = playlist_array[j];
-      playlist_array[j] = temp;
-    }
-  }
-
   /* Setup signal handlers and callbacks */
 
   ATEXIT (exit_cleanup);
@@ -387,19 +472,42 @@ int main(int argc, char **argv)
   signal (SIGCONT, signal_handler);
   signal (SIGTERM, signal_handler);
 
-  /* Play the files/streams */
-  i = 0;
-  while (i < items && !sig_request.exit) {
-    play(playlist_array[i]);
-    i++;
-  }
+  if (options.remote) {
+  
+    /* run the mainloop for the remote interface */
+    remote_mainloop();
 
+  } else {
+
+    do {
+      /* Shuffle playlist */
+      if (options.shuffle) {
+        int i;
+    
+        srandom(time(NULL));
+    
+        for (i = 0; i < items; i++) {
+          int j = i + random() % (items - i);
+          char *temp = playlist_array[i];
+          playlist_array[i] = playlist_array[j];
+          playlist_array[j] = temp;
+        }
+      }
+
+      /* Play the files/streams */
+      i = 0;
+      while (i < items && !sig_request.exit) {
+        play(playlist_array[i]);
+        i++;
+      }
+    } while (options.repeat);
+
+  }
   playlist_array_destroy(playlist_array, items);
 
-  exit (0);
+  exit (exit_status);
 }
 
-
 void play (char *source_string)
 {
   transport_t *transport;
@@ -489,9 +597,15 @@ void play (char *source_string)
 				    _("Playing: %s"), source_string);
 
   /* Skip over audio */
-  if (options.seekpos > 0.0) {
-    if (!format->seek(decoder, options.seekpos, DECODER_SEEK_START)) {
-      status_error(_("Could not skip %f seconds of audio."), options.seekpos);
+  if (options.seekoff > 0.0) {
+    /* Note: it may be simpler to handle this condition by just calling:
+     *   handle_seek_opt(&options, decoder, format);
+     * which was introduced with the remote control interface. However, that
+     * function does not call buffer_thread_kill() on error, which is
+     * necessary in this situation.
+     */
+    if (!format->seek(decoder, options.seekoff, DECODER_SEEK_START)) {
+      status_error(_("Could not skip %f seconds of audio."), options.seekoff);
       if (audio_buffer != NULL)
 	buffer_thread_kill(audio_buffer);
       return;
@@ -511,7 +625,20 @@ void play (char *source_string)
 	break;
       }
 
-      if (sig_request.pause) {
+	if (options.remote) {
+	
+		/* run the playloop for the remote interface */
+		if (remote_playloop()) {
+			/* end song requested */
+			eof = eos = 1;
+			break;
+		}
+
+		/* Skip over audio */
+		handle_seek_opt(&options, decoder, format);
+	}
+
+	if (sig_request.pause) {
 	if (audio_buffer)
 	  buffer_thread_pause (audio_buffer);
 
@@ -615,7 +742,7 @@ void play (char *source_string)
   status_message(1, _("Done."));
   
   if (sig_request.exit)
-    exit (0);
+    exit (exit_status);
 }
 
 
