@@ -6,6 +6,9 @@
  * Licensed under the GNU GPL, distributed with this program.
  */
 
+#ifdef HAVE_CONFIG_H
+#include <config.h>
+#endif
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -13,6 +16,7 @@
 #include <string.h>
 #include <stdarg.h>
 #include <getopt.h>
+#include <math.h>
 
 #include <ogg/ogg.h>
 #include <vorbis/codec.h>
@@ -25,6 +29,12 @@
 
 #define CHUNK 4500
 
+#ifdef _WIN32
+#define I64FORMAT "I64d"
+#else
+#define I64FORMAT "lld"
+#endif
+
 struct vorbis_release {
     char *vendor_string;
     char *desc;
@@ -38,7 +48,10 @@ struct vorbis_release {
         {"Xiphophorus libVorbis I 20011231", "1.0 rc3"},
         {"Xiph.Org libVorbis I 20020717", "1.0"},
         {"Xiph.Org libVorbis I 20030909", "1.0.1"},
-        {"Xiph.Org libVorbis I 20040629", "1.1.0 rc1"},
+        {"Xiph.Org libVorbis I 20040629", "1.1.0"},
+	{"Xiph.Org libVorbis I 20050304", "1.1.1"},
+	{"Xiph.Org libVorbis I 20050304", "1.1.2"},
+	{"Xiph.Org libVorbis I 20070622", "1.2.0"},
         {NULL, NULL},
     };
 
@@ -99,6 +112,8 @@ typedef struct {
     ogg_int64_t firstgranulepos;
 
     int doneheaders;
+
+    ogg_int64_t framenum_expected;
 } misc_theora_info;
 
 static int printinfo = 1;
@@ -156,7 +171,7 @@ static void error(char *format, ...)
     va_end(ap);
 }
 
-static void check_xiph_comment(stream_processor *stream, int i, char *comment,
+static void check_xiph_comment(stream_processor *stream, int i, const char *comment,
     int comment_length)
 {
     char *sep = strchr(comment, '=');
@@ -187,7 +202,7 @@ static void check_xiph_comment(stream_processor *stream, int i, char *comment,
     if(broken)
 	return;
 
-    val = comment;
+    val = (unsigned char *)comment;
 
     j = sep-comment+1;
     while(j < comment_length)
@@ -281,9 +296,29 @@ static void check_xiph_comment(stream_processor *stream, int i, char *comment,
          }
 
          if(broken) {
+             char *simple = malloc (comment_length + 1);
+             char *seq = malloc (comment_length * 3 + 1);
+             static char hex[] = {'0', '1', '2', '3', '4', '5', '6', '7', 
+                                  '8', '9', 'A', 'B', 'C', 'D', 'E', 'F'};
+             int i, c1 = 0, c2 = 0;
+             for (i = 0; i < comment_length; i++) {
+               seq[c1++] = hex[((unsigned char)comment[i]) >> 4];
+               seq[c1++] = hex[((unsigned char)comment[i]) & 0xf];
+               seq[c1++] = ' ';
+
+               if(comment[i] < 0x20 || comment[i] > 0x7D)
+                 simple[c2++] = '?';
+               else
+                 simple[c2++] = comment[i];
+             }
+             seq[c1] = 0;
+             simple[c2] = 0;
              warn(_("Warning: Illegal UTF-8 sequence in comment "
-                   "%d (stream %d): invalid sequence\n"), i, stream->num);
+                   "%d (stream %d): invalid sequence \"%s\": %s\n"), i, 
+                   stream->num, simple, seq);
              broken = 1;
+             free (simple);
+             free (seq);
              break;
          }
 
@@ -295,11 +330,12 @@ static void check_xiph_comment(stream_processor *stream, int i, char *comment,
              warn(_("Warning: Failure in utf8 decoder. This should be impossible\n"));
              return;
 	 }
+         *sep = 0;
+         if(!broken) {
+           info("\t%s=%s\n", comment, decoded);
+           free(decoded);
+         }
      }
-     
-     *sep = 0;
-     info("\t%s=%s\n", comment, decoded);
-     free(decoded);
 }
 
 static void theora_process(stream_processor *stream, ogg_page *page)
@@ -307,12 +343,21 @@ static void theora_process(stream_processor *stream, ogg_page *page)
     ogg_packet packet;
     misc_theora_info *inf = stream->data;
     int i, header=0;
+    int res;
 
     ogg_stream_pagein(&stream->os, page);
     if(inf->doneheaders < 3)
         header = 1;
 
-    while(ogg_stream_packetout(&stream->os, &packet) > 0) {
+    while(1) {
+        res = ogg_stream_packetout(&stream->os, &packet);
+        if(res < 0) {
+           warn(_("Warning: discontinuity in stream (%d)\n"), stream->num);
+           continue;
+        }
+        else if (res == 0)
+            break;
+
         if(inf->doneheaders < 3) {
             if(theora_decode_header(&inf->ti, &inf->tc, &packet) < 0) {
                 warn(_("Warning: Could not decode theora header "
@@ -353,13 +398,13 @@ static void theora_process(stream_processor *stream, ogg_page *page)
 		else
 		{
 		    float frameaspect = (float)inf->ti.frame_width/(float)inf->ti.frame_height * (float)inf->ti.aspect_numerator/(float)inf->ti.aspect_denominator; 
-		    info(_("Pixel aspect ratio %d:%d (1:%f)\n"), inf->ti.aspect_numerator, inf->ti.aspect_denominator, (float)inf->ti.aspect_numerator/(float)inf->ti.aspect_denominator);
-                    if(abs(frameaspect - 4.0/3.0) < 0.02)
+		    info(_("Pixel aspect ratio %d:%d (%f:1)\n"), inf->ti.aspect_numerator, inf->ti.aspect_denominator, (float)inf->ti.aspect_numerator/(float)inf->ti.aspect_denominator);
+                    if(fabs(frameaspect - 4.0/3.0) < 0.02)
 			info(_("Frame aspect 4:3\n"));
-                    else if(abs(frameaspect - 16.0/9.0) < 0.02)
+                    else if(fabs(frameaspect - 16.0/9.0) < 0.02)
 			info(_("Frame aspect 16:9\n"));
 		    else
-			info(_("Frame aspect 1:%d\n"), frameaspect);
+			info(_("Frame aspect %f:1\n"), frameaspect);
 		}
 
 		if(inf->ti.colorspace == OC_CS_ITU_REC_470M)
@@ -391,19 +436,37 @@ static void theora_process(stream_processor *stream, ogg_page *page)
 		}
 	    }
 	}
+        else {
+            ogg_int64_t framenum;
+            ogg_int64_t iframe,pframe;
+            ogg_int64_t gp = packet.granulepos;
+
+            if(gp > 0) {
+                iframe=gp>>inf->ti.granule_shift;
+                pframe=gp-(iframe<<inf->ti.granule_shift);
+                framenum = iframe+pframe;
+                if(inf->framenum_expected >= 0 && 
+                    inf->framenum_expected != framenum)
+                {
+                    warn(_("Warning: Expected frame %" I64FORMAT 
+                           ", got %" I64FORMAT "\n"), 
+                           inf->framenum_expected, framenum);
+                }
+                inf->framenum_expected = framenum + 1;
+            }
+            else if (inf->framenum_expected >= 0) {
+                inf->framenum_expected++;
+            }
+        }
     }
 
     if(!header) {
         ogg_int64_t gp = ogg_page_granulepos(page);
         if(gp > 0) {
             if(gp < inf->lastgranulepos)
-#ifdef _WIN32
-                warn(_("Warning: granulepos in stream %d decreases from %I64d to %I64d" ),
+                warn(_("Warning: granulepos in stream %d decreases from %" 
+                        I64FORMAT " to %" I64FORMAT "\n"),
                         stream->num, inf->lastgranulepos, gp);
-#else
-                warn(_("Warning: granulepos in stream %d decreases from %lld to %lld" ),
-                        stream->num, inf->lastgranulepos, gp);
-#endif
             inf->lastgranulepos = gp;
         }
         if(inf->firstgranulepos < 0) { /* Not set yet */
@@ -428,19 +491,11 @@ static void theora_end(stream_processor *stream)
     milliseconds = (long)((time - minutes*60 - seconds)*1000);
     bitrate = inf->bytes*8 / time / 1000.0;
 
-#ifdef _WIN32
     info(_("Theora stream %d:\n"
-           "\tTotal data length: %I64d bytes\n"
+           "\tTotal data length: %" I64FORMAT " bytes\n"
            "\tPlayback length: %ldm:%02ld.%03lds\n"
            "\tAverage bitrate: %f kb/s\n"), 
             stream->num,inf->bytes, minutes, seconds, milliseconds, bitrate);
-#else
-    info(_("Theora stream %d:\n"
-           "\tTotal data length: %lld bytes\n"
-           "\tPlayback length: %ldm:%02ld.%03lds\n"
-           "\tAverage bitrate: %f kb/s\n"), 
-            stream->num,inf->bytes, minutes, seconds, milliseconds, bitrate);
-#endif
 
     theora_comment_clear(&inf->tc);
     theora_info_clear(&inf->ti);
@@ -453,18 +508,29 @@ static void vorbis_process(stream_processor *stream, ogg_page *page )
 {
     ogg_packet packet;
     misc_vorbis_info *inf = stream->data;
-    int i, header=0;
+    int i, header=0, packets=0;
     int k;
+    int res;
 
     ogg_stream_pagein(&stream->os, page);
     if(inf->doneheaders < 3)
         header = 1;
 
-    while(ogg_stream_packetout(&stream->os, &packet) > 0) {
+    while(1) {
+        res = ogg_stream_packetout(&stream->os, &packet);
+        if(res < 0) {
+           warn(_("Warning: discontinuity in stream (%d)\n"), stream->num);
+           continue;
+        }
+        else if (res == 0)
+            break;
+
+        packets++;
         if(inf->doneheaders < 3) {
             if(vorbis_synthesis_headerin(&inf->vi, &inf->vc, &packet) < 0) {
                 warn(_("Warning: Could not decode vorbis header "
-                       "packet - invalid vorbis stream (%d)\n"), stream->num);
+                       "packet %d - invalid vorbis stream (%d)\n"), 
+                        inf->doneheaders, stream->num);
                 continue;
             }
             inf->doneheaders++;
@@ -526,17 +592,16 @@ static void vorbis_process(stream_processor *stream, ogg_page *page )
         ogg_int64_t gp = ogg_page_granulepos(page);
         if(gp > 0) {
             if(gp < inf->lastgranulepos)
-#ifdef _WIN32
-                warn(_("Warning: granulepos in stream %d decreases from %I64d to %I64d" ),
+                warn(_("Warning: granulepos in stream %d decreases from %" 
+                        I64FORMAT " to %" I64FORMAT "\n" ),
                         stream->num, inf->lastgranulepos, gp);
-#else
-                warn(_("Warning: granulepos in stream %d decreases from %lld to %lld" ),
-                        stream->num, inf->lastgranulepos, gp);
-#endif
             inf->lastgranulepos = gp;
         }
-        else {
-            warn(_("Negative granulepos on vorbis stream outside of headers. This file was created by a buggy encoder\n"));
+        else if(packets) {
+            /* Only do this if we saw at least one packet ending on this page.
+             * It's legal (though very unusual) to have no packets in a page at
+             * all - this is occasionally used to have an empty EOS page */
+            warn(_("Negative or zero granulepos (%lld) on vorbis stream outside of headers. This file was created by a buggy encoder\n"), gp);
         }
         if(inf->firstgranulepos < 0) { /* Not set yet */
         }
@@ -557,19 +622,11 @@ static void vorbis_end(stream_processor *stream)
     milliseconds = (long)((time - minutes*60 - seconds)*1000);
     bitrate = inf->bytes*8 / time / 1000.0;
 
-#ifdef _WIN32
     info(_("Vorbis stream %d:\n"
-           "\tTotal data length: %I64d bytes\n"
+           "\tTotal data length: %" I64FORMAT " bytes\n"
            "\tPlayback length: %ldm:%02ld.%03lds\n"
            "\tAverage bitrate: %f kb/s\n"), 
             stream->num,inf->bytes, minutes, seconds, milliseconds, bitrate);
-#else
-    info(_("Vorbis stream %d:\n"
-           "\tTotal data length: %lld bytes\n"
-           "\tPlayback length: %ldm:%02ld.%03lds\n"
-           "\tAverage bitrate: %f kb/s\n"), 
-            stream->num,inf->bytes, minutes, seconds, milliseconds, bitrate);
-#endif
 
     vorbis_comment_clear(&inf->vc);
     vorbis_info_clear(&inf->vi);
@@ -650,6 +707,7 @@ static void theora_start(stream_processor *stream)
 
     stream->data = calloc(1, sizeof(misc_theora_info));
     info = stream->data;
+    info->framenum_expected = -1;
 }
 
 static void vorbis_start(stream_processor *stream)
@@ -740,14 +798,22 @@ static stream_processor *find_stream_processor(stream_set *set, ogg_page *page)
         }
         else if(packet.bytes >= 7 && memcmp(packet.packet, "\x01vorbis", 7)==0)
             vorbis_start(stream);
-        else if(packet.bytes >= 7 && memcmp(packet.packet, "\x80theora", 7)==0) 
+        else if(packet.bytes >= 7 && memcmp(packet.packet, "\x80theora", 7)==0)
             theora_start(stream);
-        else if(packet.bytes >= 8 && memcmp(packet.packet, "OggMIDI\0", 8)==0) 
+        else if(packet.bytes >= 8 && memcmp(packet.packet, "OggMIDI\0", 8)==0)
             other_start(stream, "MIDI");
-        else if(packet.bytes >= 4 && memcmp(packet.packet, "fLaC", 4)==0) 
+        else if(packet.bytes >= 5 && memcmp(packet.packet, "\177FLAC", 5)==0)
             other_start(stream, "FLAC");
-        else if(packet.bytes >= 8 && memcmp(packet.packet, "Speex   ", 8)==0) 
+        else if(packet.bytes >= 8 && memcmp(packet.packet, "Speex   ", 8)==0)
             other_start(stream, "speex");
+        else if(packet.bytes >= 8 && memcmp(packet.packet, "fishead\0", 8)==0)
+            other_start(stream, "skeleton");
+        else if(packet.bytes >= 5 && memcmp(packet.packet, "BBCD\0", 5)==0)
+            other_start(stream, "dirac");
+        else if(packet.bytes >= 8 && memcmp(packet.packet, "KW-DIRAC", 8)==0)
+            other_start(stream, "dirac (old style)");
+        else if(packet.bytes >= 9 && memcmp(packet.packet, "\x80kate\0\0\0\0", 9)==0)
+            other_start(stream, "kate");
         else
             other_start(stream, NULL);
 
@@ -782,13 +848,9 @@ static int get_next_page(FILE *f, ogg_sync_state *sync, ogg_page *page,
     char *buffer;
     int bytes;
 
-    while((ret = ogg_sync_pageout(sync, page)) <= 0) {
+    while((ret = ogg_sync_pageseek(sync, page)) <= 0) {
         if(ret < 0)
-#ifdef _WIN32
-            warn(_("Warning: Hole in data found at approximate offset %I64d bytes. Corrupted ogg.\n"), *written);
-#else
-            warn(_("Warning: Hole in data found at approximate offset %lld bytes. Corrupted ogg.\n"), *written);
-#endif
+            warn(_("Warning: Hole in data (%d bytes) found at approximate offset %" I64FORMAT " bytes. Corrupted ogg.\n"), *written);
 
         buffer = ogg_sync_buffer(sync, CHUNK);
         bytes = fread(buffer, 1, CHUNK, f);
@@ -901,9 +963,14 @@ static void process_file(char *filename) {
     fclose(file);
 }
 
+static void version (void) {
+    printf (_("ogginfo from %s %s\n"), PACKAGE, VERSION);
+}
+
 static void usage(void) {
-    printf(_("ogginfo 1.1.0\n"
-             "(c) 2003-2005 Michael Smith <msmith@xiph.org>\n"
+    version ();
+    printf (_(" by the Xiph.Org Foundation (http://www.xiph.org/)\n\n"));
+    printf(_("(c) 2003-2005 Michael Smith <msmith@xiph.org>\n"
              "\n"
              "Usage: ogginfo [flags] file1.ogg [file2.ogg ... fileN.ogg]\n"
              "Flags supported:\n"
@@ -911,7 +978,9 @@ static void usage(void) {
              "\t-q Make less verbose. Once will remove detailed informative\n"
              "\t   messages, two will remove warnings\n"
              "\t-v Make more verbose. This may enable more detailed checks\n"
-             "\t   for some stream types.\n\n"));
+             "\t   for some stream types.\n"));
+    printf (_("\t-V Output version information and exit\n"));
+    /*printf (_("  -V, --version           Output version information and exit\n"));*/
 }
 
 int main(int argc, char **argv) {
@@ -931,11 +1000,14 @@ int main(int argc, char **argv) {
         exit(1);
     }
 
-    while((ret = getopt(argc, argv, "hvq")) >= 0) {
+    while((ret = getopt(argc, argv, "hqvV")) >= 0) {
         switch(ret) {
             case 'h':
                 usage();
                 return 0;
+            case 'V':
+                version();
+                return 0;
             case 'v':
                 verbose++;
                 break;
